package com.jpsoft.smart.electricity.server.codec;

import com.jpsoft.smart.electricity.server.AttributeKeys;
import com.jpsoft.smart.electricity.server.protocol.HeartbeatReceivePacket;
import com.jpsoft.smart.electricity.server.protocol.MeterReceivePacket;
import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.ByteToMessageDecoder;
import lombok.extern.slf4j.Slf4j;

import java.net.SocketAddress;
import java.util.Date;
import java.util.List;

@Slf4j
public class MeterDecodeHandler extends ByteToMessageDecoder {
    private final short HB_START_SYMBOL = 0x5E; //^
    private final short HB_END_SYMBOL = 0x24; //$
    private final short HB_BASE_LENGTH = 6;

    private final short METER_START_SYMBOL = 0x8A;
    private final short METER_END_SYMBOL = 0x66;
    private final short METER_BASE_LENGTH = 20;

    @Override
    protected void decode(ChannelHandlerContext ctx, ByteBuf byteBuf, List<Object> out) throws Exception {
        printRecvData(ctx.channel().remoteAddress(),byteBuf);

        // 屏蔽非本协议的客户端 byteBuf中保留的累计接收数据，所以取第一个一定是开头
//        if (byteBuf.getUnsignedByte(0) != HB_START_SYMBOL
//         && byteBuf.getUnsignedByte(0)  != METER_START_SYMBOL) {
//            ctx.channel().close();
//            return;
//        }

        boolean result = parseHeartbeat(ctx,byteBuf,out);
        
        if(!result){
            parseMeterPacket(ctx,byteBuf,out);
        }
    }

    /**
     * 解析电表数据包
     * @param ctx
     * @param byteBuf
     * @param out
     * @return
     */
    private boolean parseMeterPacket(ChannelHandlerContext ctx, ByteBuf byteBuf, List<Object> out) {
        boolean result = false;
        int startIndex = -1;

        byteBuf.markReaderIndex();

        while (byteBuf.readableBytes()>0) {
            if(byteBuf.readUnsignedByte() == METER_START_SYMBOL){
                startIndex = byteBuf.readerIndex()-1;
                break;
            }
        }

        // 截止符可能在数据段中重复出现
//        while (byteBuf.readableBytes()>0) {
//            if(byteBuf.readUnsignedByte() == METER_END_SYMBOL){
//                endIndex = byteBuf.readerIndex()-1;
//                break;
//            }
//        }

        byteBuf.resetReaderIndex();

        //累计接收数据保护完整数据段
        if (startIndex +  METER_BASE_LENGTH <= byteBuf.readableBytes()){
            log.info("找到完整数据段!");

            byte[] buffer = new byte[METER_BASE_LENGTH];

            byteBuf.getBytes(startIndex,buffer);

            MeterReceivePacket meterPacket = new MeterReceivePacket();

            Date requestTime = (Date)ctx.channel().attr(AttributeKeys.REQUEST_TIME).get();
            meterPacket.setRequestTime(requestTime);

            meterPacket.setResponseTime(new Date());

            meterPacket.setMachineNo(((buffer[2] & 0xff) << 8) + (buffer[1] & 0xff));

            meterPacket.setUnitNo(((buffer[4] & 0xff) << 8) + (buffer[3] & 0xff));

            meterPacket.setHouseNo(buffer[5] & 0xff);

            //当前状态
            meterPacket.setState(buffer[8] & 0xff);

            //剩余电量
            meterPacket.setRemain(((buffer[9] & 0xff) << 8) + (buffer[10] & 0xff));

            //购电次数
            meterPacket.setRechargeTimes(((buffer[11] & 0xff) << 8) + (buffer[12] & 0xff));

            //验证码1
            int valCode1 = buffer[16] & 0xff;

            int cmpValCode1 = buffer[8] & 0xff;

            for (int i = 9; i <= 15; i++)
            {
                cmpValCode1 ^= (buffer[i] & 0xff);
            }

            if (valCode1 == cmpValCode1)
            {
                meterPacket.setValidateResult(true);
            }
            else
            {
                meterPacket.setValidateResult(false);
            }

            out.add(meterPacket);

            byteBuf.skipBytes(byteBuf.readableBytes());

            result = true;
        }

        return result;
    }

    /**
     * 解析心跳信号
     * @param ctx
     * @param byteBuf
     * @param out
     * @return
     */
    private boolean parseHeartbeat(ChannelHandlerContext ctx, ByteBuf byteBuf, List<Object> out) {
        boolean result = false;
        int startIndex = -1;
        int endIndex = -1;

        byteBuf.markReaderIndex();

        while (byteBuf.readableBytes()>0) {
            if(byteBuf.readUnsignedByte() == HB_START_SYMBOL){
                startIndex = byteBuf.readerIndex()-1;
                break;
            }
        }

        while (byteBuf.readableBytes()>0) {
            if(byteBuf.readUnsignedByte() == HB_END_SYMBOL){
                endIndex = byteBuf.readerIndex()-1;
                break;
            }
        }

        byteBuf.resetReaderIndex();

        if (endIndex - startIndex + 1 == HB_BASE_LENGTH){
            log.info("找到完整心跳数据包!");

            byte[] buffer = new byte[HB_BASE_LENGTH];

            byteBuf.getBytes(startIndex,buffer);

            HeartbeatReceivePacket hrp = new HeartbeatReceivePacket();

            StringBuilder sb = new StringBuilder();

            for (int i=0;i<buffer.length;i++){
                sb.append((char)buffer[i]);
            }

            hrp.setTerminalNo(sb.substring(1,5));

            out.add(hrp);

            byteBuf.skipBytes(byteBuf.readableBytes());

            result = true;
        }

        return result;
    }

    private void printRecvData(SocketAddress remoteAddress, ByteBuf byteBuf) {
        //打印接收到的数据
        byteBuf.markReaderIndex();

        StringBuilder sb = new StringBuilder();

        while (byteBuf.readableBytes()>0) {
            if (sb.length()>0){
                sb.append(" ");
            }

            short b = byteBuf.readUnsignedByte();

            //java中byte是-128~127，C#中是0~255，需要转化
            String str = Integer.toHexString(b);

            if (str.length()==1){
                sb.append("0");
            }

            sb.append(str.toUpperCase());
        }

        log.warn(String.format("收到来自%s数据(累计):%s",remoteAddress.toString(), sb.toString()));

        byteBuf.resetReaderIndex();
    }
}
